World [
	Title:	"World/view for GARP2500"
	Author:	"John Niclasen"
	Purpose: {
		Implement SDL2 / OpenGL / GLSL based View extension to World.
	}
]


if value? 'GARPview [exit]


GARPview: make context! [


V-BOX-TYPE:			1
V-BUTTON-TYPE:		2
V-DIAL-TYPE:		3
V-DRAGGER-TYPE:		4
V-FIELD-TYPE:		5
V-JACK-TYPE:		6
V-PANEL-TYPE:		7
V-ROTARY-TYPE:		8
V-SCROLLER-TYPE:	9
V-SENSOR-TYPE:		10
V-SLIDER-TYPE:		11
V-STEP-DIAL-TYPE:	12
V-TOGGLE-TYPE:		13
V-VBOX-TYPE:		14
V-STYLE-TYPE:		15


N-STYLES:	32
styles: []
vertex-shaders: []
fragment-shaders: []
shader-programs: []
textures: []


; --- Worldview Object ---
full-vob: make struct! [[
	pointer			pane
	pointer			vobs
	pointer			parent
	pointer			data
	pointer			over
	pointer			engage
	uint32			type		; vob type
	sint32			x			; x position
	sint32			y			; y position
	sint32			w			; width
	sint32			h			; height
	uint32			pane-size	; # entries in pane
	uint32			nvobs		; # vobs
	uint32			rgba		; # Red-Green-Blue-Alpha
	uint32			shader-program
	uint32			texture		; texture ID (GLuint)
	sint32			u			; texture u coord
	sint32			v			; texture v coord
	sint32			tw			; texture width
	sint32			th			; texture height
	uint32			rgba-loc	; rgba variable location
	uint32			tex-loc		; texture variable location
	;uint32			model-loc	; model variable location
	;uint32			cam-loc		; camera variable location
	;uint32			proj-loc	; projection variable location
	uint32			mvp-loc		; model view projection variable location
	uint32			VAO			; Vertex Array Object
	uint32			VBO			; Vertex Buffer Object
	uint32			show		; Show the vob or not
	uint32			user-data	; Free for use (May be pointer later)
	uint32			state		; Object state
	uint32			style		; Vob style
	uint32		filler0

	uint32		filler1
	uint32		filler2
	uint32		filler3
	uint32		filler4
	uint32		filler5
	uint32		filler6
	uint32		filler7
	uint32		filler8
	] []
]
vob: make struct! [[
	pointer			pane
	pointer			vobs
	pointer			parent
	pointer			data
	pointer			over
	pointer			engage
	uint32			type		; vob type
	sint32			x			; x position
	sint32			y			; y position
	sint32			w			; width
	sint32			h			; height
	uint32			pane-size	; # entries in pane
	uint32			nvobs		; # vobs
	uint32			rgba		; # Red-Green-Blue-Alpha
	uint32			shader-program
	uint32			texture		; texture ID (GLuint)
	sint32			u			; texture u coord
	sint32			v			; texture v coord
	sint32			tw			; texture width
	sint32			th			; texture height
	uint32			rgba-loc	; rgba variable location
	uint32			tex-loc		; texture variable location
	;uint32			model-loc	; model variable location
	;uint32			cam-loc		; camera variable location
	;uint32			proj-loc	; projection variable location
	uint32			mvp-loc		; model view projection variable location
	uint32			VAO			; Vertex Array Object
	uint32			VBO			; Vertex Buffer Object
	uint32			show		; Show the vob or not
	uint32			user-data	; Free for use (May be pointer later)
	uint32			state		; Object state
	uint32			style		; Vob style
	uint32		filler0
	] []
]


parse-specs: make function! [[
	specs	[block!]
	ptr		[handle!]
	vobs	[vector!]
	/debug
	/local
		rules vob-rules coords texture-rules box-rules button-rules dial-rules dragger-rules
		field-rules jack-rules
		panel-rules rotary-rules scroller-rules sensor-rules slider-rules step-dial-rules toggle-rules vbox-rules
		style-rules
		vob-init fill-worldview-object
		pos
		; helper variables
		value v1 v2 begin range cx cy steps
		x y w h tw th
		rgba
		v-shader f-shader texture
		id user-data style
		bus
		worldview-object
		; Worldview Object
		;full-vob vob
		button dial dragger jack
		rotary step-dial toggle vbox
][
	; --- Worldview Object ---
	button: copy vob/1
	append button [
		uint32			vLoc
	]
	button: make struct! reduce [button []]


	dial: copy vob/1
	append dial [
		float			range-begin
		float			range-end
		sint32			cx
		sint32			cy
		float			angle
	]
	dial: make struct! reduce [dial []]


	dragger: copy vob/1
	append dragger [
		sint32			orig-x
		sint32			orig-y
		sint32			dx
		sint32			dy
		sint32			xrange
		sint32			yrange
	]
	dragger: make struct! reduce [dragger []]


	jack: copy vob/1
	append jack [
		;uint32			v-offsetLoc
		uint32			bus
		;uint32			RGBLoc
	]
	jack: make struct! reduce [jack []]


	rotary: copy vob/1
	append rotary [
		uint32			v-offsetLoc
	]
	rotary: make struct! reduce [rotary []]


	step-dial: copy vob/1
	append step-dial [
		float			range-begin
		float			range-end
		uint32			steps
		sint32			cx
		sint32			cy
		float			angle
		float			step-angle
	]
	step-dial: make struct! reduce [step-dial []]


	toggle: copy vob/1
	append toggle [
		uint32			vLoc
	]
	toggle: make struct! reduce [toggle []]


	vbox: copy vob/1
	append vbox [
		uint32			vLoc
		float			v-offset
	]
	vbox: make struct! reduce [vbox []]


	; --- Rules ---
	coords: [
		set value pair! (v1: value/x v2: value/y)
		| [set v1 number! set v2 number!]
	]


	texture-rules: [
		; word! will not work, as next worldview element is a word
		;set value word! (probe value: do bind value specs)
		set value file! (
			texture: to lit-string! value
		) coords (
			x: v1
			y: v2
		)
		coords (
			tw: v1
			th: v2
		)
	]


	vob-init: [ (
		w: h: v-shader: f-shader: texture: x: y: tw: th: none
		rgba: 0
		;user-data: 0
		style: bus: 0
	) ]


	vob-rules: [
		set value pair! (
			either w [
				make error! append copy "Syntax error: " value
			][
				w: value/x h: value/y
			]
		)
		| set value number! (
			either w [
				either h [
					make error! append copy "Syntax error: " value
				][
					h: value
				]
			][
				w: value
			]
		)
		| set value tuple! (
			rgba: to integer! to binary! value
		)
		| 'vertex-shader (if debug [prin " vertex-shader"]) set value file! (
			if v-shader [make error! append copy "Syntax error: " value]
			if word! = type? value [value: do bind value specs]
			v-shader: to lit-string! value
		)
		| 'fragment-shader (if debug [prin " fragment-shader"]) set value file! (
			if f-shader [make error! append copy "Syntax error: " value]
			if word! = type? value [value: do bind value specs]
			f-shader: to lit-string! value
		)
		| 'texture (if debug [prin " texture"]) texture-rules
		| 'user-data (if debug [prin " user-data"]) set value integer! (user-data: value)
		| 'bus (if debug [prin " bus"]) set value integer! (bus: value)
	]


	fill-worldview-object: [
		worldview-object/x: pos/1
		worldview-object/y: pos/2
		worldview-object/w: either w w [either tw tw 100]
		worldview-object/h: either h h [either th th 100]
		worldview-object/rgba: rgba
		either texture [
			either id: find textures texture [
				worldview-object/texture: index? id
			][
				append textures texture
				worldview-object/texture: length? textures
			]
			worldview-object/u: either x x 0
			worldview-object/v: either y y 0
			worldview-object/tw: either tw tw 1
			worldview-object/th: either th th 1
		][
			worldview-object/texture: 0
		]
		either v-shader [
			either id: find vertex-shaders v-shader [
				v-shader: index? id
			][
				append vertex-shaders v-shader
				v-shader: length? vertex-shaders
			]
			either f-shader [
				either id: find fragment-shaders f-shader [
					f-shader: index? id
				][
					append fragment-shaders f-shader
					f-shader: length? fragment-shaders
				]
			][
				f-shader: 0		; Default fragment shader
			]
			either id: find/only shader-programs reduce [v-shader f-shader] [
				worldview-object/shader-program: index? id
			][
				append/only shader-programs retain reduce [v-shader f-shader]
				worldview-object/shader-program: length? shader-programs
			]
		][
			v-shader: 0							; Default vertex shader
			worldview-object/shader-program: 0	; Default shader program
		]
		worldview-object/user-data: user-data
		user-data: 0
		worldview-object/style:		style
	]


	box-rules: [
		vob-init
		any [
			vob-rules
		] (
			either vobs [
				worldview-object: as vob vobs/:user-data
			][
				worldview-object: ptr	; jack
			]
			worldview-object/type: V-BOX-TYPE
			do fill-worldview-object
			pos/2: pos/2 + worldview-object/h
		)
	]


	button-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as button vobs/:user-data
			worldview-object/type: V-BUTTON-TYPE
			do fill-worldview-object
			worldview-object/state: 0
			pos/2: pos/2 + worldview-object/h
		)
	]


	dial-rules: [
		vob-init (begin: range: 0.0 cx: cy: none)
		set begin real!
		set range real!
		opt [set cx real! set cy real!]
		any [
			vob-rules
		] (
			worldview-object: as dial vobs/:user-data
			worldview-object/type: V-DIAL-TYPE
;tau:		6.283185307179586	; 2 * pi
			worldview-object/range-begin: begin / 360.0 * 6.283185307179586
			worldview-object/range-end: begin + range / 360.0 * 6.283185307179586
			do fill-worldview-object
			worldview-object/cx: either cx cx [to integer! worldview-object/w / 2]
			worldview-object/cy: either cy cy [to integer! worldview-object/h / 2]
			worldview-object/angle: 0.0
			pos/2: pos/2 + worldview-object/h
		)
	]


	dragger-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as dragger vobs/:user-data
			worldview-object/type: V-DRAGGER-TYPE
			do fill-worldview-object
			worldview-object/orig-x: worldview-object/x
			worldview-object/orig-y: worldview-object/y
			worldview-object/dx: 0
			worldview-object/dy: 0
			worldview-object/xrange: 0
			worldview-object/yrange: 541
			pos/2: pos/2 + worldview-object/h
		)
	]


	field-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as vob vobs/:user-data
			worldview-object/type: V-FIELD-TYPE
			do fill-worldview-object
			pos/2: pos/2 + worldview-object/h
		)
	]


	jack-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as jack vobs/:user-data
			worldview-object/type: V-JACK-TYPE
			do fill-worldview-object
			worldview-object/bus: bus
			pos/2: pos/2 + worldview-object/h
		)
	]


	panel-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: ptr	; TODO as panel ?
			worldview-object/type: V-PANEL-TYPE
			do fill-worldview-object
			pos/1: pos/2: 0
		)
		into rules
	]


	rotary-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as rotary vobs/:user-data
			worldview-object/type: V-ROTARY-TYPE
			do fill-worldview-object
			worldview-object/state: 0
			pos/2: pos/2 + worldview-object/h
		)
	]


	scroller-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as vob vobs/:user-data
			worldview-object/type: V-SCROLLER-TYPE
			do fill-worldview-object
			pos/2: pos/2 + worldview-object/h
		)
	]


	sensor-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as vob vobs/:user-data
			worldview-object/type: V-SENSOR-TYPE
			do fill-worldview-object
			pos/2: pos/2 + worldview-object/h
		)
	]


	slider-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as vob vobs/:user-data
			worldview-object/type: V-SLIDER-TYPE
			do fill-worldview-object
			;print ["slider x:" worldview-object/x "y:" worldview-object/y]
			pos/2: pos/2 + worldview-object/h
		)
	]


	step-dial-rules: [
		vob-init (begin: range: 0.0 steps: 0 cx: cy: none)
		set begin real!
		set range real!
		set steps integer!
		opt [set cx real! set cy real!]
		any [
			vob-rules
		] (
			worldview-object: as step-dial vobs/:user-data
			worldview-object/type: V-STEP-DIAL-TYPE
;tau:		6.283185307179586	; 2 * pi
			worldview-object/range-begin: begin / 360.0 * 6.283185307179586
			worldview-object/range-end: begin + range / 360.0 * 6.283185307179586
			worldview-object/steps: steps
			do fill-worldview-object
			worldview-object/cx: either cx cx [to integer! worldview-object/w / 2]
			worldview-object/cy: either cy cy [to integer! worldview-object/h / 2]
			worldview-object/angle: 0.0
			worldview-object/step-angle: 0.0
			pos/2: pos/2 + worldview-object/h
		)
	]


	toggle-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as toggle vobs/:user-data
			worldview-object/type: V-TOGGLE-TYPE
			do fill-worldview-object
			worldview-object/state: 0
			pos/2: pos/2 + worldview-object/h
		)
	]


	vbox-rules: [
		vob-init
		any [
			vob-rules
		] (
			worldview-object: as vbox vobs/:user-data
			worldview-object/type: V-VBOX-TYPE
			do fill-worldview-object
			;worldview-object/state: 0
			pos/2: pos/2 + worldview-object/h
		)
	]


	style-rules: [
		set value word! (
			if debug [prin ["^/" value]]
			if not find styles value [make error! append copy "Unknown word or style: " value]
		)
		vob-init
		(style: index? find styles value)
		any [
			vob-rules
		] (
			either bus = 0 [
				worldview-object: as vob vobs/:user-data
				worldview-object/type: V-STYLE-TYPE	; no type; take from style
				do fill-worldview-object
			][
				worldview-object: as jack vobs/:user-data
				worldview-object/type: V-STYLE-TYPE	; no type; take from style
				do fill-worldview-object
				worldview-object/bus: bus
			]
			pos/2: pos/2 + worldview-object/h
		)
	]


	rules: [
		any [
			set value set-word! (
				if debug [prin ["^/" value]]
				append styles to word! value
				user-data: length? styles
			)
			| 'at (if debug [prin "^/at"]) [
				set value pair! (pos/1: value/x pos/2: value/y)
				| set x number! set y number! (pos/1: x pos/2: y)
			]
			| 'box (if debug [prin "^/box"]) box-rules
			| 'button (if debug [prin "^/button"]) button-rules
			| 'dial (if debug [prin "^/dial"]) dial-rules
			| 'dragger (if debug [prin "^/dragger"]) dragger-rules
			| 'field (if debug [prin "^/field"]) field-rules
			| 'jack (if debug [prin "^/jack"]) jack-rules
			| 'panel (if debug [prin "^/panel"]) panel-rules
			| 'rotary (if debug [prin "^/rotary"]) rotary-rules
			| 'scroller (if debug [prin "^/scroller"]) scroller-rules
			| 'sensor (if debug [prin "^/sensor"]) sensor-rules
			| 'slider (if debug [prin "^/slider"]) slider-rules
			| 'step-dial (if debug [prin "^/step-dial"]) step-dial-rules
			| 'toggle (if debug [prin "^/toggle"]) toggle-rules
			| 'vbox (if debug [prin "^/vbox"]) vbox-rules
			| style-rules
		]
		end
	]


	; --- Main ---
	pos:		[0 0]
	pos/1:
	pos/2:		0
	user-data:	0
	parse specs rules
]]


set 'GARPview-layout make function! [[
	specs	[block!]
	ptr		[handle!]
	/debug
	/local vobs
][
	ptr: as vob ptr
	either ptr/pane-size > 0 [
		vobs: as reduce ['vector! full-vob ptr/pane-size] ptr/vobs
	][
		vobs: none
	]
	either debug [
		parse-specs/debug specs ptr vobs
	][
		parse-specs specs ptr vobs
	]
]]


set 'GARPview-stylize make function! [[
    specs	[block!]	;'"A block of: new-style: old-style facets"
	ptr		[handle!]
	/debug
	/local vobs
][
	vobs: as reduce ['vector! full-vob N-STYLES] ptr
	either debug [
		parse-specs/debug specs ptr vobs
	][
		parse-specs specs ptr vobs
	]
]]


]	; GARPview context


