World [
	Title:		"Cortex Preferences"
	Date:		12-Dec-2022
	Version:	1.0.0
	History: [
		1.0.0	[12-12-2022	JN	{Special version for G2500}]
		0.8.2	[11-2-2018	JN	{Improved tab-completion for worlds holding values of type file!
								 Improved HELP of file! datatype
								 Added EXIT to INCLUDE}]
	]
]

any-block!:		make typeset! [block! paren! path! set-path! get-path! lit-path!]
\{
any-context!:	make typeset! [context! error! port!]
any-function!:	make typeset! [operator! function! routine! task!]
;any-object!:	make typeset! [context! error! port!]
any-paren!:		make typeset! [paren!]
any-path!:		make typeset! [path! set-path! get-path! lit-path!]
}
any-string!:	make typeset! [string! binary! file! email! url! tag!]
\{
any-type!:		make typeset! [unset! none! logic! integer! real! complex! percent! char! pair! range! time! date! string! lit-string! binary! file! email! url! tag! issue! bitset! tuple! vector! image! block! list! paren! path! set-path! get-path! lit-path! map! datatype! typeset! word! set-word! get-word! lit-word! refinement! operator! function! routine! callback! context! error! task! task-id! port! handle! struct! library! node! comment! KWATZ!]
any-word!:		make typeset! [word! set-word! get-word! lit-word! refinement!]
immediate!:		make typeset! [unset! none! logic! integer! real! complex! percent! char! pair! range! time! date! lit-string! issue! tuple! datatype! typeset! word! set-word! get-word! lit-word! refinement! task-id!]
}
number!:		make typeset! [integer! real! percent!]
\{
scalar!:		make typeset! [integer! real! complex! percent! char! pair! range! tuple! time!]
}
series!:		make typeset! [string! binary! file! email! url! tag! block! list! paren! path! set-path! get-path! lit-path! KWATZ!]

\{
e:			2.718281828459045
pi:			3.141592653589793
tau:		6.283185307179586	; 2 * pi
inf:		1.0 / 0.0
-inf:		-1.0 / 0.0
nan:		1.0 / 0.0 * 0.0

null:		#"^@"
esc:		#"^["
bs:			#"^H"
cr:			#"^M"
lf:			#"^/"
newline:	#"^/"
newpage:	#"^L"
tab:		#"^-"
slash:		#"/"
backslash:	#"\"

crlf:		"^M^/"
}

; Control
for: make function! [[
	'"Evaluate a block over a range of values."
	[throw retain]
	'word [word!] '"Variable to hold current value"
	start [number! char!] '"Starting value"
	end [number! char!] '"Ending value"
	bump [number! char!] '"Amount to skip each time"
	body [block!] '"Block to evaluate each time"
	/local do-body
][
	;if end == #"^(FF)" [
		;make error! join "FOR - invalid argument: " end
	;]
	do-body: make function! reduce [reduce [[throw retain] word] body]
	start: start - bump
	either bump < 0 [
		while [end <= start: start + bump] [
			do-body start
		]
	][
		while [end >= start: start + bump] [
			do-body start
		]
	]
]]
foreach: make function! [[
	'"Evaluate a block for each value(s) in a series."
	[throw retain]
	'word [word! block!] '"Word or block or words to set each time"
	data [series! map!] '"Series to traverse"
	body [block!] '"Block to evaluate each time"
	/local c l body' result
][
	if map! = type? data [data: to block! data]
	if 0 < length? data [
		c: copy [none]
		either block! = type? word [
			l: length? word
			while [l > 0] [
				insert c to set-word! word/:l
				l: l - 1
			]
			c: make context! bind c body
			body': bind copy body c
			l: length? word
			set/local/at word data c
			result: do body'
			\{
			data': copy/shallow data
			while [0 < length? data': skip data' l] [
				set/local/at word data' c
				result: do body'
			]
			}
			while [0 < length? data: skip data l] [
				set/local/at word data c
				result: do body'
			]
			set/local/at word none c
		][
			insert c to set-word! word
			c: make context! bind c body
			body': bind copy body c
			set/local/at word data/1 c
			result: do body'
			;data': copy/shallow data
			while [0 < length? data: next data] [
				set/local/at word data/1 c
				result: do body'
			]
			set/local/at word none c
		]
	]
	result
]]
func: make function! [[
	'"Define a function with given spec and body."
	[retain]
	spec [block!] '"^"Description^" followed by arguments (opt type and string)"
	body [block!] '"The body block of the function"
][
	make function! reduce [spec body]
]]
\{
loop: make function! [[
	'"Evaluate a block a specified number of times."
	[throw retain]
	count [integer!] '"Number of repetitions"
	block [block!] '"Block to evaluate each time"
][
	;if false = compiled? block [
		;compile/at block 'loop
		;compile/at block block
		;compile block
	;]
	while [0 <= count: count - 1] block
]]
}
repeat: make function! [[
	'"Evaluate a block a number of times."
	[throw retain]
	'word [word!] '"Word to set each time"
	value [integer!] '"Maximum number"
	body [block!] '"Block to evaluate each time"
	/local spec do-body start
][
	spec: [[throw retain]]
	append spec word
	do-body: make function! reduce [spec body]
	remove/last spec
	start: 0
	while [value > start] [
		start: start + 1
		do-body start
	]
]]
switch: make function! [[
	'"Select a choice and evaluate the block that follows it."
	[throw retain]
	value '"Value to search for"
	cases [block!] '"Block of cases to search"
	/default
		case '"Default case if no others are found"
	/local to-do
][
	to-do: find cases either datatype! = type? :value [
		to word! value
	][
		:value
	]
	either to-do [
		while [all [to-do: find next to-do block! block! <> type? pick to-do 1]] []
		either block! = type? pick to-do 1 [
			do pick to-do 1
		][
			if default [do case]
		]
	][
		if default [do case]
	]
]]
until: make function! [[
	'"Evaluate a block until it is true."
	[throw retain]
	block [block!]
	/local result
][
	while [not result: do block] []
	result
]]

; Datatype
error?:			make function! [['"True for error values." value][error! = type? :value]]

; Help
probe: make function! [[
	'"Print a molded value and return that same value."
	;[retain]
	value
][
	print mold :value
	:value
]]

; Math
max: make function! [[
	'"The greater of two values."
	value1
	value2
][
	;pick reduce [value1 value2] value1 > value2
	if :value1 > :value2 [return :value1] :value2
]]
\{
min: make function! [[
	'"The lesser of two values."
	value1
	value2
][
	;pick reduce [value1 value2] value1 < value2
	if :value1 < :value2 [return :value1] :value2
]]
}

; Port, File and IO
load: make function! [[
	'"Load a file or string. Bind block to global context."
	[retain]
	source	[file! string! lit-string! url!]
	/library '"Force file to be a dynamic library"
	/lines '"Handle data as lines"
	/all '"Also load potential header and empty lines (if /lines)"
	/local data
][
	data: switch/default type? :source [
		string!
		lit-string!	[source]
		file!	[
			either library [
				return load-library source
			][
				either all [
					as string! read/binary source
				][
					read source
				]
			]
		]
		url!	[read source]
	] [
		make error! reduce ['script 'invalid-arg mold :source]
	]
	either lines [
		data: parse-utils/split-string data make bitset! "^/"
		if empty? pick data: tail data -1 [
			remove data: back data
		]
		data: head data
		either all [
			while [0 < length? data] [
				change/only data make block! data/1
				set-newline data true
				data: next data
			]
		][
			while [0 < length? data] [
				change/only data to block! data/1
				either 0 = length? data/1 [
					remove data
				][
					set-newline data true
					data: next data
				]
			]
		]
		data: head data
		set-newline 'data true
	][
		data: either all [
			make block! data
		][
			to block! data
		]
	]
	if not any [	; NOT ANY because of /all
		all
		'World <> pick data 1
		block! <> type? pick data 2
	] [
		remove/part data 2
	]
	bind data 'system
]]
save: make function! [[
	'"Save to a file"
	;[retain]
	where [file!] '"Where to save it"
	value '"Value to save"
	/lines '"Handle block of data as lines"
	/local s
][
	either block! = type? :value [
		;value: copy/shallow value
		;s: clear ""
		s: copy ""
		while [0 < length? value] [
			append s either newline? value [
				#"^/"
			][
				#" "
			]
			append s either lines [
				form value/1
			][
				mold value/1
			]
			value: next value
		]
		write where next s
	][
		write where mold :value
	]
]]

; Series
join: make function! [[
	'"Concatenate values."
	[retain]
	value '"Base value"
	rest '"Value or block of values"
][
	value: either find series! type? :value [copy value] [form :value]
	append value reduce :rest
]]
sort: make function! [[
	'"Sort a series."
	series [series!]
	;/skip '"Treat the series as records of fixed size"
	;size [integer!] '"Size of each record"
][
	\{
	either skip [
		if (length? series) // size <> 0 [
			make error! [script out-of-range]
		]
		sys-utils/qsort-size series 1 (length? series) / size size
	][
	}
		sys-utils/qsort series 1 length? series
	;]
	series
]]

parse-utils: make context! [

bparse: make function! [[
	'"Parse a block according to rules."
	[retain]
	input [any-block!] '"Input block to parse"
	rules [block!] '"Rules to parse by"
	g [context!]
	/case '"Use case-sensitive comparison"
	/local keyword-rules type-rules
	i j n keyword token var value dest status mark rules'
][
	;rules': copy/shallow rules
	rules': rules
	keyword-rules: [
		any [
			until [
				not switch/default type?/word :token type-rules [	; default
					either (case and (token == input/1))
							or ((not case) and (token = input/1)) [
						input: next input
						true
					][
						false
					]
				]
			]
			true
		]
		end [0 = length? input]
		into [
			either find any-block! type? input/1 [
				either block! = type? :token [
					insert/only g/istack input
					insert g/istack g/level
					g/level: 0
					insert/only g/istack input/1
					status: bparse input/1 token g
					g/level: g/istack/1
					remove g/istack
					if status [
						input: g/istack/1
						input: next input
					]
					remove g/istack
					status
				][
					clear g/istack
					make error! join '"PARSE - invalid argument: " :token
				]
			][
				false
			]
		]
		opt [
			switch/default type?/word :token type-rules [	; default
				if (case and (token == input/1))
						or ((not case) and (token = input/1)) [
					input: next input
				]
			]
			true
		]
		skip [
			either 0 = length? input [
				false
			][
				input: next input
				true
			]
		]
		some [
			if status: switch/default type?/word :token type-rules [	; default
				either (case and (token == input/1))
						or ((not case) and (token = input/1)) [
					input: next input
					true
				][
					false
				]
			] [
				until [
					not switch/default type?/word :token type-rules [	; default
						either (case and (token == input/1))
								or ((not case) and (token = input/1)) [
							input: next input
							true
						][
							false
						]
					]
				]
				status: true
			]
			status
		]
		thru [
			either token = 'end [
				false
			][
				either none = dest: find/only input token [
					false
				][
					;input next dest
					input: skip input 1 + (index? dest) - index? input
					true
				]
			]
		]
		to [
			either token = 'end [
				input: tail input
				true
			][
				either none = dest: find/only input token [
					false
				][
					;input: dest
					input: skip input (index? dest) - index? input
					true
				]
			]
		]
		| [
			rules': none
			true
		]
	]

	type-rules: [
		none! [
			true
		]
		block! [
			insert/only g/istack input
			;either bparse copy/shallow input token g
			either bparse input token g [
				input: g/istack/1
				remove g/istack
				true
			][
				false
			]
		]
		paren! [
			do token
			true
		]
		datatype! [
			either all [
				find to typeset! token type? input/1
				0 < length? input
			] [
				input: next input
				true
			][
				false
			]
		]
		typeset! [
			either all [
				find token type? input/1
				0 < length? input
			] [
				input: next input
				true
			][
				false
			]
		]
	]

	;input: skip input 0
	;insert/only g/istack skip input 0
	;insert/only g/istack copy/shallow input

	g/level: g/level + 1
	mark: none

	status: true
	while [all [rules' 0 < length? rules']] [
		i: j: keyword: none
		;get-token
		token: rules'/1
		rules': next rules'
		if word! = type? :token [
			if not find [any copy end into opt set skip some thru to |] token [
				token: get/at token rules
			]
		]	; get-token end
		switch type?/word :token [
			integer! [
				i: max 0 token
				;get-token
				token: rules'/1
				rules': next rules'
				if word! = type? :token [
					if not find [any copy end into opt set skip some thru to |] token [
						token: get/at token rules
					]
				]	; get-token end
				switch type?/word :token [
					none! [
						clear g/istack
						make error! join '"PARSE - unexpected end of rule after: " i
					]
					integer! [
						j: token
						;get-token
						token: rules'/1
						rules': next rules'
						if word! = type? :token [
							if not find [any copy end into opt set skip some thru to |] token [
								token: get/at token rules
							]
						]	; get-token end
						switch type?/word :token [
							none! [
								clear g/istack
								make error! join '"PARSE - unexpected end of rule after: " j
							]
							word! [
								either find [end skip |] token [
									keyword: token
								][
									if find [any into opt some thru to] token [
										keyword: token
										;get-token
										token: rules'/1
										rules': next rules'
										if word! = type? :token [
											if not find [any copy end into opt set skip some thru to |]
													token [
												token: get/at token rules
											]
										]	; get-token end
									]
								]
							]
						]
					]
					word! [
						either find [end skip |] token [
							keyword: token
						][
							if find [any into opt some thru to] token [
								keyword: token
								;get-token
								token: rules'/1
								rules': next rules'
								if word! = type? :token [
									if not find [any copy end into opt set skip some thru to |] token [
										token: get/at token rules
									]
								]	; get-token end
							]
						]
					]
				]
			]
			get-word! [
				input: get/at to word! token rules
			]
			set-word! [
				;set/local/at token copy/shallow input rules
				set/local/at token input rules
			]
			word! [
				switch token [
					any into opt some thru to [
						keyword: token
						;get-token
						token: rules'/1
						rules': next rules'
						if word! = type? :token [
							if not find [any copy end into opt set skip some thru to |] token [
								token: get/at token rules
							]
						]	; get-token end
					]
					end skip | [keyword: token]
					copy [
						keyword: token
						var: rules'/1
						rules': next rules'
						if word! <> type? :var [
							clear g/istack
							make error! join '"PARSE - invalid argument: " :var
						]
					]
					set [
						keyword: token
						var: rules'/1
						rules': next rules'
						if word! <> type? :var [
							clear g/istack
							make error! join '"PARSE - invalid argument: " :var
						]
						;get-token
						token: rules'/1
						rules': next rules'
						if word! = type? :token [
							if not find [any copy end into opt set skip some thru to |] token [
								token: get/at token rules
							]
						]	; get-token end
					]
				]
			]
		]
		either i [
			either j [
				status: true
				n: i
				while [all [status n > 0]] [
					either keyword [
						status: switch keyword keyword-rules
					][
						status: switch/default type?/word :token type-rules [	; default
							either (case and (token == input/1))
									or ((not case) and (token = input/1)) [
								input: next input
								true
							][
								false
							]
						]
					]
					n: n - 1
				]
				if status [
					n: j - i
					while [all [status n > 0]] [
						either keyword [
							status: switch keyword keyword-rules
						][
							status: switch/default type?/word :token type-rules [	; default
								either (case and (token == input/1))
										or ((not case) and (token = input/1)) [
									input: next input
									true
								][
									false
								]
							]
						]
						n: n - 1
					]
					status: true
				]
			][
				status: true
				while [all [status i > 0]] [
					either keyword [
						status: switch keyword keyword-rules
					][
						status: switch/default type?/word :token type-rules [	; default
							either (case and (token == input/1))
									or ((not case) and (token = input/1)) [
								input: next input
								true
							][
								false
							]
						]
					]
					i: i - 1
				]
			]
			if mark [
				if mark =? input = false [
					set/local/at var copy/part mark input rules
				]
				mark: none
			]
		][
			either keyword [
				switch/default keyword [
					copy [
						; 9-Feb-2018 mark: copy/shallow input
						mark: input
						status: true
					]
					set [
						value: input/1
						status: switch/default type?/word :token type-rules [	; default
							either (case and (token == input/1))
									or ((not case) and (token = input/1)) [
								input: next input
								true
							][
								false
							]
						]
						if status [set/local/at var value rules]
					]
				][
					if any [
						all [
							not find [end skip |] keyword
							word! = type? :token
							find [any copy into opt set some thru to |] :token
						]
						all [
							not find [end thru to] keyword
							:token == 'end
						]
					] [
						clear g/istack
						make error! join '"PARSE - invalid argument: " :token
					]
					status: switch keyword keyword-rules
					if mark [
						if mark =? input = false [
							set/local/at var copy/part mark input rules
						]
						mark: none
					]
				]
			][
				either all [
						datatype! <> type? :token
						find [get-word! set-word!] type?/word :token
				] [
					status: true
				][
					status: switch/default type?/word :token type-rules [	; default
						either (case and (token == input/1))
								or ((not case) and (token = input/1)) [
							input: next input
							true
						][
							false
						]
					]
				]
				if mark [
					if mark =? input = false [
						set/local/at var copy/part mark input rules
					]
					mark: none
				]
			]
		]
		if not status [
			; 9-Feb-2018 input: copy/shallow g/istack/1
			input: g/istack/1
			if all [rules' rules': find rules' '|] [
				if 0 = length? rules': next rules' [
					status: true
				]
			]
		]
	]
	g/level: g/level - 1
	either g/level > 0 [
		either status [
			;dest: g/istack/1
			;dest: skip dest (index? input) - index? dest
			insert/only remove g/istack input
		][
			remove g/istack
		]
		return status
	][
		remove g/istack
		either status and (0 = length? input) [
			return true
		][
			return false
		]
	]
]]

set 'parse make function! [[
	'"Parse a series according to rules."
	[retain]
	input [series!] '"Input series to parse"
	rules [block! string! none!] '"Rules to parse by"
	/all '"Parse all chars including spaces"
	/case '"Use case-sensitive comparison"
	/local g
][
	if not find [block! string! none!] type?/word rules [
		make error! "parse expected rules argument of type: block! string! none!"
	]
	switch/default type?/word input [
		block! paren! path! [
			g: make context! [
				;istack: reduce [copy/shallow input]
				istack: make block! 4
				level: 0
			]
			;insert/only g/istack copy/shallow input
			insert/only g/istack input
			either case [
				;bparse/case copy/shallow input rules g
				bparse/case input rules g
			][
				;bparse copy/shallow input rules g
				bparse input rules g
			]
		]
\{
		string! binary! file! email! url! tag! KWATZ! [
			g: make context! [
				;istack: reduce [as string! input]
				istack: make block! 4
				level: 0
			]
			insert g/istack as string! input
			either all [
				either case [
					sparse/all/case as string! copy input rules g
				][
					sparse/all as string! copy input rules g
				]
			][
				either case [
					sparse/case as string! copy input rules g
				][
					sparse as string! copy input rules g
				]
			]
		]
}
	] [
		make error! "parse expected input argument of type: series!"
	]
]]
]	; context

; Strings
form: make function! [[
	'"Convert a value into a human-readable string."
	[retain]
	value '"The value to form"
	/local result
][
	switch/default type?/word :value [
		string! [
			;copy/shallow value
			value
		]
		block! paren! map! [
			result: make string! 8
			if 0 < length? value [
				;value: either map! = type? value [
					;to block! value
				;][
					;copy/shallow value
				;]
				if map! = type? value [
					value: to block! value
				]
				insert result form value/1
				value: next value
				while [0 < length? value] [
					result: tail result
					insert result either newline? value [
						#"^/"
					][
						#" "
					]
					result: next result
					insert result form value/1
					value: next value
				]
			]
			result: head result
		]
		none! tag! path! set-path! get-path! lit-path! set-word! get-word! lit-word! issue! refinement! [
			mold value
		]
		binary! [
			result: mold value
			remove/part result 2
			remove/last result
			result
		]
		datatype! [
			result: to string! value
			if #"!" = pick result length? result [
				remove/last result
			]
			result
		]
	] [
		to string! :value
	]
]]

; System

sys-utils: make context! [

; ---- print-last-error ----
print-last-error: make function! [[
	/local str err id
][
	if system/last-error/task-id <> 0 [
		prin '"^/** Message from task: "
		prin system/last-error/task-id
		prin '" "
		print system/last-error/task-name
	]
	err: select system/errors system/last-error/type
	id: select err system/last-error/id
	str: append copy "** " reduce [
		err/type
		'": "
		either block! = type? id [
			form reduce compile/at id system/last-error
		][
			id
		]
	]
	if 74 < length? str [str: head append clear skip str 74 '"..."]
	print str
	str: append copy "** Near: " system/last-error/near
	if 74 < length? str [str: head append clear skip str 74 '"..."]
	print str
]]


; ---- qsort ----
qsort: make function! [[
	;[retain]
	v [series!]
	left [integer!]
	right [integer!]
	/local last i j t u
][
	if left >= right [exit]
	j: to integer! left + right / 2
	t: pick v left
	poke v left pick v j
	poke v j t
	last: left
	i: left + 1
	while [i <= right] [
		t: pick v i
		u: pick v left
		either (type? t) = type? u [
			if t < u [
				next 'last
				u: pick v last
				poke v last t
				poke v i u
			]
		][
			if (type? t) < type? u [
				next 'last
				u: pick v last
				poke v last t
				poke v i u
			]
		]
		next 'i
	]
	t: pick v left
	poke v left pick v last
	poke v last t
	qsort v left last - 1
	qsort v last + 1 right
]]


]; sys-utils

do %user.w

